#!/usr/bin/env python
'''
  #############################################################################
  #
  #     This Cplant(TM) source code is the property of Sandia National
  #     Laboratories.
  #
  #     This Cplant(TM) source code is copyrighted by Sandia National
  #     Laboratories.
  #
  #     The redistribution of this Cplant(TM) source code is subject to the
  #     terms of the GNU Lesser General Public License
  #     (see cit/LGPL or http://www.gnu.org/licenses/lgpl.html)
  #
  #     Cplant(TM) Copyright 1998, 1999, 2000, 2001, 2002 Sandia Corporation.
  #     Under the terms of Contract DE-AC04-94AL85000, there is a non-exclusive
  #     license for use of this work by or on behalf of the US Government.
  #     Export of this program may require a license from the United States
  #     Government.
  #
  #############################################################################

 File    : testbase.py
 Revision: $Revision: 1.6 $
 Author  : William McLendon, Narayan Desai
 Description:
	implementation of base test classes.
	Contains the following classes:
		Node
		Msgbuf
		Person
		TestInfo
		TODependency
                TODependencies
                boolOR
                boolAND
                boolNAND
                boolNOR
		tOutput
		tInput
		Test
'''

import re
from time import sleep
from string import join
from xml.dom.minidom import parse,Document
from exceptions import Exception
from errorhandlers import TestSetupError

class Node:
    '''This is an easier interface to DOM objects
    self.typemap is a dispatch table for children
    self.__attrfields__ are the attributes to display
    self.__textfields__ are textnodes to treat like attributes'''
    typemap={}
    __attrfields__=[]
    __textfields__=[]
    def __init__(self,xml):
        self.tagname=xml.tagName
        self.children=[]
        self.__setdefaults__()
        map(lambda (attr,value):setattr(self,attr,value),xml._get_attributes().items())
        for node in filter(lambda x:x.nodeType==1,xml.childNodes):
            if node.tagName in self.__textfields__:
                setattr(self,node.tagName,join(map(lambda x:x.nodeValue,filter(lambda x:x.nodeType==3,node.childNodes)),""))
                continue
            if not self.typemap.has_key(node.tagName):
                self.children.append(Node(node))
            else:
                self.children.append(self.typemap[node.tagName](node))

    def __show__(self):
        for key in dir(self):
            print key, getattr(self,key)

    def toxmldom(self):
        ret=Document().createElement(self.tagname)
        for attr in self.__attrfields__:
            ret.setAttribute(attr,getattr(self,attr))
        for child in self.children:
            ret.appendChild(child.toxmldom())
        for text in self.__textfields__:
            tn=Document().createElement(text)
            tn.appendChild(Document().createTextNode(join(getattr(self,text),'')))
            ret.appendChild(tn)
        return ret

    def __str__(self):
        return self.tagname+join(map(lambda x:"%s:%s"%(x,getattr(self,x)), dir(self)),' ')

    def __setdefaults__(self):
        pass

    def Match(self,spec):
        return 1
        for attr in [x for x in dir(spec) if x[0] != '_']:
            if (getattr(spec,attr) != '*') and  getattr(self,attr) != getattr(spec,attr):
                return 0
        for childtype in spec._children.keys():
            if childtype not in self._children.keys():
                return 0
            required=map(lambda x:x, spec._children[childtype])
            for node in required:
                if filter(lambda x:node._Match(x),self._children[childtype]):
                    required.remove(node)
            if required:
                return 0
        return 1

class Msgbuf(Node):
    def __init__(self,xml):
        self.tagname=xml.tagName
        self.__setdefaults__()
        map(lambda (attr,value):setattr(self,attr,value),xml._get_attributes().items())
        self.__value__=xml.childNodes[0].nodeValue

    def toxmldom(self):
        ret=Document().createElement(self.tagname)
        map(lambda x:ret.setAttribute(x,getattr(self,x)), self.__attrfields__)
        tn=Document().createTextNode(self.__value__)
        ret.appendChild(tn)
        return ret

class Person(Node):
    __attrfields__=['name','email','phone']
    def __setdefaults__(self):
        self.email="N/A"
        self.name ="N/A"
        self.phone="N/A"

class TestInfo(Node):
    typemap={'author':Person,'contact':Person}
    __textfields__=['description']

# -----------------------------------------------------------------------------
# TestOptions Objects
# -----------------------------------------------------------------------------
class TODependency(Node):
    ''' TestOptions.dependency '''
    __attrfields__=['type','name','minPctMatch','maxPctMatch']
    def __setdefaults__(self):
        self.name = None
        self.type = "PREREQ"
        self.minPctMatch = u'0.0'		# Default
        self.maxPctMatch = u'100.0'		# Default

# wcm: not sure how to define this , but it seems to work ATM
class boolOR(Node):
  __attrfields__=[]
class boolNOR(Node):
  __attrfields__=[]
class boolAND(Node):
  __attrfields__=[]
class boolNAND(Node):
  __attrfields__=[]
class boolNOT(Node):
  __attrfields__=[]
boolAND.typemap= {'dependency':TODependency, 'AND':boolAND, 'OR':boolOR, 'NAND':boolNAND, \
                  'NOR':boolNOR, 'NOT':boolNOT}
boolOR.typemap=  {'dependency':TODependency, 'AND':boolAND, 'OR':boolOR, 'NAND':boolNAND, \
                  'NOR':boolNOR, 'NOT':boolNOT}
boolNOR.typemap= {'dependency':TODependency, 'AND':boolAND, 'OR':boolOR, 'NAND':boolNAND, \
                  'NOR':boolNOR, 'NOT':boolNOT}
boolNAND.typemap={'dependency':TODependency, 'AND':boolAND, 'OR':boolOR, 'NAND':boolNAND, \
                  'NOR':boolNOR, 'NOT':boolNOT}
boolNOT.typemap= {'dependency':TODependency, 'AND':boolAND, 'OR':boolOR, 'NAND':boolNAND, \
                  'NOR':boolNOR, 'NOT':boolNOT}

class TODependencies(Node):
  ''' dependency listing'''
  typemap={'dependency':TODependency, 'AND':boolAND, 'NAND':boolNAND, 'OR':boolOR, \
           'NOR':boolNOR,'NOT':boolNOT}

class tOutput(Msgbuf):
    ''' output from tested component '''
    __attrfields__=['type', 'format']

class tInput(Msgbuf):
    ''' input sources '''
    __attrfields__=['type']

# ---------------------------------------------------------------------------------
# Abstract Test Class
# ---------------------------------------------------------------------------------
class Test(Node):
    ''' Test -- abstract class from which all tests should inherit '''
    __attrfields__=['name','numReps','preDelay','postDelay','iterDelay','minPctMatch','maxPctMatch','onMismatch']
    typemap = {'dependencies':TODependencies}
    
    def setup(self): raise NotImplementedError,"test.setup(self)"
    def execute(self, scratch): raise NotImplementedError,"test.execute(self, scratch)"

    def __init__(self,xml):
        # Flags
        # Initialize ATTRIBUTES
        self.irep    = 0		# actual reps tried
        self.numReps = '1'		# total reps to do
        self.preDelay = '0.0'
        self.postDelay = '0.0'
        self.iterDelay = '0.0'
        self.minPctMatch = '0.0'
        self.maxPctMatch = '100.0'
        self.onMismatch = 'CONTINUE'        # [CONTIUE],BREAK,EXIT

        # Data Members
        self.status      = None		# This test status
        self.nMatch      = 0		# number of tests where expected = actual
        self.pctMatched  = 0.0
        self.expect      = {}		# expected return value K: tagname V: buffer
        self.response    = {}           # storage for execution response K: tagname
        self.xml=xml
        Node.__init__(self,xml)

    def reset(self):
        self.irep   = 0
        self.status = None
        self.nMatch = 0
        self.pctMatched = 0.0
        self.expect.clear()
        self.response.clear()

    def Display(self):
        print "name=   \t", `self.name`
        print "nMatch= \t", `self.nMatch`
        print 'expect= \t', `self.expect`
        print 'attrs = \t', `self.__attrfields__`

    def prepare(self, options,testdata):
        '''
        Abstract test preparer.
        '''
        self.testdata = testdata
        self.reset()
        self.options = options
        self.numReps   = int(self.numReps)
        self.preDelay  = float(self.preDelay)
        self.postDelay = float(self.postDelay)
        self.iterDelay = float(self.iterDelay)
        self.minPctMatch = float(self.minPctMatch)
        self.maxPctMatch = float(self.maxPctMatch)
        for child in self.children:
            if child.tagname == 'output':
                if self.expect.has_key(child.type): 
                    self.expect[child.type].append((child.format,child.__value__))
                else:
                    self.expect[child.type] = [(child.format,child.__value__)]
        self.setup()

    def GetInput(self,itype):
        '''returns a list of input strings.  It is up to the test to decide what to do with them.'''
        r = []
        inputs=filter(lambda x:x.tagname=='input',self.children)
        src=filter(lambda x:x.type==itype,inputs)
        for i in range(len(src)):
            r.append(src[i].__value__)
        return r

    def GetOutput(self,itype):
        '''returns a list of tuples: (BUFFERTYPE,BUFFER), one entry per <output> element 
           containing an attribute that has type=itype '''
        r=[]
        outputs=filter(lambda x:x.tagname=='output',self.children)
        src=filter(lambda x:x.type==itype,outputs)
        for i in range(len(src)):
            r.append((src[i].format,src[i].__value__))
        return r
  
    def Validate(self):
        '''determines whether or not a test matches it's expected output(s)
           returns 1 if matches, 0 if not'''
        if self.options.has_key('verbose'): print "V:\tValidating test results."
        matched = 1
        self.imatch={}
        for key in self.expect.keys():
            self.imatch[key] = 'NO'
            if self.response.has_key(key):
                if self.options.has_key('verbose'):
                    print "V:\tChecking '%s' via %s"%(key,self.expect[key][0][0])
                    print "V:\t\tActual Buffer = '%s'"%(self.response[key])
                    print "V:\t\tMatch Pattern = '%s'"%(self.expect[key][0][1])
                if self.expect[key][0][0] == 'REGEXP':
                    rmatch = re.compile(self.expect[key][0][1], re.M|re.S)
                    if rmatch.match(str(self.response[key])):
                        self.imatch[key]='YES'
                elif self.expect[key][0][0] == "TXTSTR":
                    if self.expect[key][0][1] == str(self.response[key]):
                        self.imatch[key]='YES'
                    if self.options.has_key('verbose'):
                        print "V:\t\tDoes buffer match?  %s"%(self.imatch[key])
            else:
                print 60*"*"
                print " ERROR: Unrecognized source value '%s' in <response>"%(key)
                print 60*"*"
                matched = 0
        for key in self.imatch:
            if self.imatch[key] == 'NO':
                matched = 0
        return matched

    def runTest(self,scratch):
        matched = 0
        sleep(self.preDelay)
        for iter in range(self.irep, self.numReps):
            self.irep += 1
            self.testdata.nIterExec += 1

            self.execute(scratch)
            matched = self.Validate()
            self.nMatch += matched

            self.pctMatched = self.nMatch / self.numReps*100
            if self.minPctMatch <= self.pctMatched and self.pctMatched <= self.maxPctMatch:
                self.status = 'PASS'
            else:
                self.status = 'FAIL'

            # If test failed, we need to decide what to do based on test.onMismatch
            if self.options.has_key('debug'): print "D:\texec status=",self.status
            if matched == 0:
                for k in self.imatch.keys():
                    self.testdata.mismatches.append( (self.irep,self.expect[k],self.response[k]) )
#                self.testdata.mismatches.append( (self.irep) )
                if self.onMismatch == 'BREAK':
                    raise TestSetupError,"break"
                    break
                elif self.onMismatch == 'HALT':
                    raise TestSetupError,"halt"
                    exit(200)
#            print self.expect
#            print self.response

            self.response.clear()       # clear the test's response buffer(s)

            if iter < self.numReps-1:
                sleep( self.iterDelay )

    def printSummary(self, message):
        if self.status == 'FAIL':
          message = "m[%.1f%% : %.1f%%] "%(self.minPctMatch, self.minPctMatch) + message
        print "[%d of %d]\t%-10s\t %6.2f%%\t%4s     \t%s"%(self.irep,self.numReps,self.name,self.pctMatched,self.status,message)

        sleep( self.postDelay )


def dbWrite(dbTable, data):
    pass


# EOF #
