#!/usr/bin/env python

from __future__ import nested_scopes

from elementtree.ElementTree import Element, XML, tostring
from cPickle import dumps, loads, Pickler, Unpickler
from getopt import getopt, GetoptError
from xml.parsers.expat import ExpatError
from os import close,fork
from select import select
from signal import signal, SIGINT
from sys import argv, exc_info, exit
from syslog import syslog, LOG_INFO, LOG_ERR
from time import sleep, strftime
from re import compile
from traceback import extract_tb
from xml.sax import SAXParseException

from sss.restriction import Data, DataSet, ID, SizedSet
from sss.server import Server, BufferedPipe
from sss.ssslib import comm_lib, ProtocolError, CommunicationError, ConnectError, InternalError, LookupError

class Event(Data):
    pass

class Subscription(Data):
    '''Subscription handles event manager subscriptions'''
    regex_re = compile("r/(.*)/")

    def __init__(self,element, id):
        self.element = element
        self.id = id
        self.regex={}
        for (field,val) in element.attrib.iteritems():
            ere = self.regex_re.match(val)
            if ere: self.regex[field] = compile(ere.group(1))
        for (field, val) in [('matches','0'),('delivered','0'),('failures','0'), ('last', 'never')]:
            if not self.element.attrib.get(field):
                self.element.attrib[field] = val

    def EventMatch(self,event):
        for field in ['component','msg','data']:
            if self.regex.has_key(field):
                if not self.regex[field].match(event.element.attrib[field]):
                    return False
            else:
                if ((self.element.attrib[field] != event.element.attrib[field]) and (self.element.attrib[field] != "**")):
                    return False
        self.element.attrib['matches'] = str(int(self.element.attrib['matches'])+1)
        return True

class emng(Server):
    __implementation__='emng'
    __component__='event-manager'
    __statefields__ = ['subscriptions']
    __dispatch__ = {'get-event':'cache.Get',
                    'add-event':'XAddEvent',
                    'add-subscription':'subscriptions.Add',
                    'get-subscription':'subscriptions.Get',
                    'del-subscription':'subscriptions.Del'}
    
    def __setup__(self):
        self.subscriptions = DataSet('subscriptions','subscription',Subscription,ID(),True)
        self.cache = SizedSet('events', 'event',Event,None,False,600,2000)

    def XAddEvent(self, xml, (peer,port)):
        return self.cache.Add(xml, (peer,port), self.AddEvent)

    def __setupChildren__(self):
        # self.edeliver.(wr) are for events that need propagation
        self.edeliver = BufferedPipe()
        self.pipes.append(self.edeliver)
        # self.estats.(wr) are for statistics
        self.estats = BufferedPipe()
        self.pipes.append(self.estats)
        pid = fork()
        if pid == 0:
            self.todo={}
            signal(SIGINT, self.ChildSigHand)
            close(self.estats.r)
            self.readsockets=[self.edeliver.r]
            self.socketdispatch={self.edeliver.r:self.DeliverEvents}
            self.ServeForever()
        else:
            self.readsockets.append(self.estats.r)
            self.socketdispatch[self.estats.r] = self.HandleStatMsg
            close(self.edeliver.r)
            return [pid]

    def ChildSigHand(self,signum,frame):
        exit(0)

    def AddEvent(self,event, arg):
        for s in self.subscriptions:
            if s.EventMatch(event):
                self.edeliver.WriteMessage((s.element.attrib['notify'],event.Fetch(),s.id))

    def HandleStatMsg(self,pipe):
        try:
            msg=self.estats.ReadMessage()
        except OSError:
            syslog(LOG_ERR, "got broken pipe from child")
            syslog(LOG_ERR, "Statistics are offline")
            self.shut=1
        if msg.has_key('ssubids'):
            for s in self.subscriptions:
                sea = s.element.attrib
                if s.id in msg['ssubids']:
                    sea['delivered'] = str(int(sea['delivered'])+msg['ssubids'].count(s.id))
                    sea['last'] = strftime("%X %x")
        if msg.has_key('fsubids'):
            for s in self.subscriptions:
                sea = s.element.attrib
                if s.id in msg['fsubids']:
                    sea['failures'] = str(int(sea['failures'])+msg['fsubids'].count(s.id))
        if msg.has_key('rsubids'):
            for s in self.subscriptions:
                sea = s.element.attrib
                if s.id in msg['rsubids']:
                    syslog(LOG_ERR, "Subscription for %s removed because of component death"%(sea['notify']))
                    self.subscriptions.data.remove(s)
        return 0
        
    def DeliverEvents(self,pipe,inl=None):
        inl=[pipe]
        while inl:
            (dest,event,subid)=self.edeliver.ReadMessage()
            if self.todo.has_key(dest):
                self.todo[dest].append((event,subid))
            else:
                self.todo[dest]=[(event,subid)]
            # catch any other events coming down the pipe now
            (inl,foo,foo)=select([pipe],[],[],0)
        # Now we deliver contents of self.todo
        for component in self.todo.keys():
            stats={}
            try:
                msg = Element("events")
                h=self.comm.ClientInit(component)
                events=map(lambda x:msg.append(x[0]),self.todo[component])
                self.comm.SendMessage(h,tostring(msg))
                resp=self.comm.RecvMessage(h)
                self.comm.ClientClose(h)
                r = XML(resp)
                if r.tag != 'event-ok':
                    syslog(LOG_ERR, "Failed to send event to %s"%(component))
                else:
                    stats['ssubids']=[x[1] for x in self.todo[component]]
                    del self.todo[component]
                            
            except (ExpatError, SAXParseException):
                syslog(LOG_ERR, "Bad response from %s"%(component))
                stats['fsubids']=[x[1] for x in self.todo[component]]
            except LookupError:
                syslog(LOG_ERR, "Component %s down. flushing queue"%(component))
                stats['rsubids']=[x[1] for x in self.todo[component]]
                del self.todo[component]
            except ConnectError:
                syslog(LOG_ERR, "Component %s down(connection refused). flushing queue"%(component))
                stats['rsubids']=[x[1] for x in self.todo[component]]
                del self.todo[component]
            except CommunicationError,m:
                syslog(LOG_ERR, "Component %s transmit failure: %s."%(component,m))
            except:
                (t,v,tb)=exc_info()
                syslog(LOG_ERR, "Error occurred while communication with %s"%(component))
                for line in extract_tb(tb):
                    syslog(LOG_ERR, '  File "%s", line %i, in %s\n    %s\n'%line)
                    syslog(LOG_ERR, "%s: %s\n"%(t,v))
                del t,v,tb
                
            if stats: self.estats.WriteMessage(stats)
        return 0
        
if __name__ == '__main__':
    try:
        (opts,args)=getopt(argv[1:],'d',['daemon='])
    except GetoptError,msg:
        print "%s\nUsage:\nemng.py [-d] [--daemon <pidfile>]"%(msg)
        exit(1)
    daemon=filter(lambda x:(x[0] == '--daemon'),opts)
    ldebug=len(filter(lambda x:(x[0] == '-d'),opts))
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0][1])
    s=emng(debug=ldebug)
    s.ServeForever()
