#include <Python.h>
#include <ssslib.h>
#include <stdio.h>
#include <pthread.h>

staticforward PyTypeObject SSS_Type;

/* I'm going to have these exception types:

SSS_InitError
SSS_ParseError
SSS_ProtocolError
SSS_SocketError
SSS_ConnectError
SSS_CommunicationError
SSS_InternalError
SSS_LookupError

SSS_UnhandledError

*/

/* comm_lib object */

static PyObject *SSS_ParseError;
static PyObject *SSS_ProtocolError;
static PyObject *SSS_SocketError;
static PyObject *SSS_ConnectError;
static PyObject *SSS_CommunicationError;
static PyObject *SSS_InternalError;
static PyObject *SSS_InitError;
static PyObject *SSS_LookupError;
static PyObject *SSS_UnhandledError;

typedef struct {
  PyObject_HEAD
  SSS sss;
  pthread_mutex_t mutex;
} comm_lib;

/* util function for setting exceptions */

void set_sss_exception(int errnum){
  int typenum;
  char buf[MAXERRLENGTH];

  SSS_strerror(errnum,buf,&typenum);

  switch (typenum) {
  case SSS_PARSE_ERROR:
    PyErr_SetString(SSS_ParseError,buf);
    break;
  case SSS_INIT_ERROR:
    PyErr_SetString(SSS_InitError,buf);
    break;
  case SSS_PROTOCOL_ERROR:
    PyErr_SetString(SSS_ProtocolError,buf);
    break;
  case SSS_SOCKET_ERROR:
    PyErr_SetString(SSS_SocketError,buf);
    break;
  case SSS_CONNECT_ERROR:
    PyErr_SetString(SSS_ConnectError,buf);
    break;
  case SSS_COMMUNICATION_ERROR:
    PyErr_SetString(SSS_CommunicationError,buf);
    break;
  case SSS_INTERNAL_ERROR:
    PyErr_SetString(SSS_InternalError,buf);
    break;
  case SSS_LOOKUP_ERROR:
    PyErr_SetString(SSS_LookupError,buf);
    break;
  case SSS_UNKNOWN_ERROR:
    PyErr_SetString(SSS_UnhandledError,buf);
    break;
  }
}

/* constructor */

static PyObject *comm_lib_NEW(char *conffile,int debug){
  comm_lib *obj;

  obj = PyObject_NEW(comm_lib,&SSS_Type);
  if (obj)
    obj->sss = SSS_Init(conffile,debug);

  if (!obj->sss){
    /* no SSS is contructed yet, the init bombed out early, so we rely on errno */
    PyErr_SetFromErrno(SSS_InitError);
    return NULL;
  }

  /*pthread_mutex_init(&obj->mutex,NULL);*/
  
  return (PyObject *)obj;
}

static PyObject *get_config_dict(PyObject *self,PyObject *args){
  int i;
  SSS sss;
  SSS_ConfigDict dict;
  PyObject *retdict;
  PyObject *prefs;
  char **cprefs;
  /*pthread_mutex_t *mutex;*/

  sss = ((comm_lib *)self)->sss;
  /*mutex = &((comm_lib *)self)->mutex;*/

  /*pthread_mutex_lock(mutex);*/
  
  dict = SSS_GetGlobalConfigDict(sss);

  retdict = PyDict_New();
  prefs = PyList_New(0);

  for (i=0;dict->keywords[i];i++)
    PyDict_SetItem(retdict,PyString_FromString(dict->keywords[i]),
                   PyString_FromString(dict->values[i]));

  SSS_FreeConfigDict(dict);

  cprefs = SSS_GetProtocolPreference(sss);
  for (i=0;cprefs[i];i++)
    PyList_Append(prefs,PyString_FromString(cprefs[i]));

  PyDict_SetItemString(retdict,"prefs",prefs);

  /*pthread_mutex_unlock(mutex);*/
  
  return retdict;
}
  
static PyObject *comm_lib_sdregister(PyObject *self,PyObject *args){
  int retval;
  SSS sss;
  char *service;
  char *host;
  int port;
  char *protocol;
  char *schema_version;
  char *tier;
  PyObject *retobj;
  /*pthread_mutex_t *mutex;*/
  
  if (!PyArg_ParseTuple(args,"ssisss",&service,&host,&port,&protocol,&schema_version,&tier)){
    return NULL;
  }
    
  sss = ((comm_lib *)self)->sss;
  /*mutex = &((comm_lib *)self)->mutex;*/

  /*pthread_mutex_lock(mutex);*/
  
  Py_BEGIN_ALLOW_THREADS

  retval = SSS_SDRegister(sss,service,host,port,protocol,schema_version,tier);

  Py_END_ALLOW_THREADS

  /*pthread_mutex_unlock(mutex);*/
  
  if (retval == -1){
    set_sss_exception(sss->sss_errno);
    return NULL;
  }

  retobj = Py_BuildValue("i",retval);
  
  return retobj;
}

static PyObject *comm_lib_sendrecv(PyObject *self,PyObject *args){
  int retval;
  SSS sss;
  char *message;
  char *handle;
  char *msg;

  if (!PyArg_ParseTuple(args,"ss",&handle,&message))
    return NULL;

  sss = ((comm_lib *)self)->sss;

  Py_BEGIN_ALLOW_THREADS

  retval = SSS_SendMessage(sss,handle,message);

  if (retval == -1){
    set_sss_exception(sss->sss_errno);
    return NULL;
  }

  msg = SSS_RecvMessage(sss,handle);

  Py_END_ALLOW_THREADS

  if (!msg){
    if (sss->sss_errno == SSS_EOF)
      return Py_BuildValue("s","");
    set_sss_exception(sss->sss_errno);
    return NULL;
  }

  return Py_BuildValue("s",msg);
}

static PyObject *comm_lib_sdremove(PyObject *self,PyObject *args){
  int retval;
  SSS sss;
  char *service;
  char *host;
  int port;
  char *protocol;
  PyObject *retobj;
  /*pthread_mutex_t *mutex;*/
  
  if (!PyArg_ParseTuple(args,"ssis",&service,&host,&port,&protocol)){
    return NULL;
  }
    
  sss = ((comm_lib *)self)->sss;
  /*mutex = &((comm_lib *)self)->mutex;*/

  /*pthread_mutex_lock(mutex);*/

  Py_BEGIN_ALLOW_THREADS

  retval = SSS_SDRemove(sss,service,host,port,protocol);

  Py_END_ALLOW_THREADS
  
  /*pthread_mutex_unlock(mutex);*/

  if (retval == -1){
    set_sss_exception(sss->sss_errno);
    return NULL;
  }

  retobj = Py_BuildValue("i",retval);
  
  return retobj;
}

static PyObject *comm_lib_serverinit(PyObject *self, PyObject *args){
  char *retval;
  PyObject *socket;
  int sockfd;
  SSS sss;
  PyObject *retobj;
  /*pthread_mutex_t *mutex;*/

  if (!PyArg_ParseTuple(args,"O",&socket)){
    return NULL;
  }
  
  sockfd = PyObject_AsFileDescriptor(socket);

  sss = ((comm_lib *)self)->sss;
  /*mutex = &((comm_lib *)self)->mutex;*/

  /*pthread_mutex_lock(mutex);*/
  
  Py_BEGIN_ALLOW_THREADS

  retval = SSS_ServerInit(sss,sockfd);

  Py_END_ALLOW_THREADS

  /*pthread_mutex_unlock(mutex);*/

  if (!retval){
    set_sss_exception(sss->sss_errno);
    return NULL;
  }
  
  retobj = Py_BuildValue("s",retval);

  return retobj;
}

static PyObject *comm_lib_clientinit(PyObject *self,PyObject *args){
  char *retval;
  char *service;
  SSS sss;
  PyObject *retobj;
  /*pthread_mutex_t *mutex;*/

  if (!PyArg_ParseTuple(args,"s",&service)){
    return NULL;
  }

  sss = ((comm_lib *)self)->sss;
  /*mutex = &((comm_lib *)self)->mutex;*/

  /*pthread_mutex_lock(mutex);*/

  Py_BEGIN_ALLOW_THREADS

  retval = SSS_ClientInit(sss,service);

  Py_END_ALLOW_THREADS
  
  /*pthread_mutex_unlock(mutex);*/

  if (!retval){
    set_sss_exception(sss->sss_errno);
    return NULL;
  }
    
  retobj = Py_BuildValue("s",retval);

  return retobj;
}

static PyObject *comm_lib_serverclose(PyObject *self, PyObject *args){
  int retval;
  char *handle;
  SSS sss;
  PyObject *retobj;
  /*pthread_mutex_t *mutex;*/

  if (!PyArg_ParseTuple(args,"s",&handle)){
    return NULL;
  }

  sss = ((comm_lib *)self)->sss;
  /*mutex = &((comm_lib *)self)->mutex;*/

  /*pthread_mutex_lock(mutex);*/

  Py_BEGIN_ALLOW_THREADS

  retval = SSS_ServerClose(sss,handle);

  Py_END_ALLOW_THREADS
    
  /*pthread_mutex_unlock(mutex);*/

  if (retval == -1){
    set_sss_exception(sss->sss_errno);
    return NULL;
  }

  retobj = Py_BuildValue("i",retval);

  return retobj;
}

static PyObject *comm_lib_clientclose(PyObject *self,PyObject *args){
  int retval;
  char *handle;
  SSS sss;
  PyObject *retobj;
  /*pthread_mutex_t *mutex;*/

  if (!PyArg_ParseTuple(args,"s",&handle)){
    return NULL;
  }

  sss = ((comm_lib *)self)->sss;
  /*mutex = &((comm_lib *)self)->mutex;*/

  /*pthread_mutex_lock(mutex);*/

  Py_BEGIN_ALLOW_THREADS

  retval = SSS_ClientClose(sss,handle);

  Py_END_ALLOW_THREADS

  /*pthread_mutex_unlock(mutex);*/

  if (retval == -1){
    set_sss_exception(sss->sss_errno);
    return NULL;
  }
  
  retobj = Py_BuildValue("i",retval);

  return retobj;
}

static PyObject *comm_lib_sendmessage(PyObject *self,PyObject *args){
  int retval;
  char *handle;
  char *message;
  SSS sss;
  PyObject *retobj;
  /*pthread_mutex_t *mutex;*/

  if (!PyArg_ParseTuple(args,"ss",&handle,&message)){
    return NULL;
  }

  sss = ((comm_lib *)self)->sss;
  /*mutex = &((comm_lib *)self)->mutex;*/

  /*pthread_mutex_lock(mutex);*/

  Py_BEGIN_ALLOW_THREADS

  retval = SSS_SendMessage(sss,handle,message);

  Py_END_ALLOW_THREADS

  /*pthread_mutex_unlock(mutex);*/

  if (retval == -1){
    set_sss_exception(sss->sss_errno);
    return NULL;
  }

  retobj = Py_BuildValue("i",retval);

  return retobj;
}

static PyObject *comm_lib_recvmessage(PyObject *self,PyObject *args){
  char *retval;
  char *handle;
  SSS sss;
  PyObject *retobj;
  /*pthread_mutex_t *mutex;*/

  if (!PyArg_ParseTuple(args,"s",&handle)){
    return NULL;
  }

  sss = ((comm_lib *)self)->sss;
  /*mutex = &((comm_lib *)self)->mutex;*/

  /*pthread_mutex_lock(mutex);*/

  Py_BEGIN_ALLOW_THREADS

  retval = SSS_RecvMessage(sss,handle);

  Py_END_ALLOW_THREADS

  /*pthread_mutex_unlock(mutex);*/

  if (!retval){
    if (sss->sss_errno == SSS_EOF){
      return Py_BuildValue("s","");
    }
    set_sss_exception(sss->sss_errno);
    return NULL;
  }

  if (retval)
    retobj = Py_BuildValue("s",retval);
  else
    retobj = Py_BuildValue("s","");

  /* free(retval);*/
  
  return retobj;
}
  
static PyObject *set_servprot(PyObject *self,PyObject *args){
  char *p;
  SSS sss;
  
  if (!PyArg_ParseTuple(args,"s",&p))
    return NULL;

  sss = ((comm_lib *)self)->sss;

  SSS_SetServerProtocol(sss,p);
  return Py_None;
}

static PyObject *static_lookup(PyObject *self,PyObject *args){
  char *s;
  SSS sss;
  SSS_ServiceList ret;
  PyObject *retdict;

  if (!PyArg_ParseTuple(args,"s",&s))
    return NULL;

  sss = ((comm_lib *)self)->sss;

  /* FIXME - only returns first element */

  ret = SSS_StaticLookup(sss,s);

  if (!ret){
    set_sss_exception(sss->sss_errno);
    return NULL;
  }

  retdict = PyDict_New();

  PyDict_SetItem(retdict,PyString_FromString("protocol"),
                 PyString_FromString(ret->services[0]->protocol));
  PyDict_SetItem(retdict,PyString_FromString("service"),
                 PyString_FromString(ret->services[0]->service));
  PyDict_SetItem(retdict,PyString_FromString("extra"),
                 PyString_FromString(ret->services[0]->extra));
  PyDict_SetItem(retdict,PyString_FromString("host"),
                 PyString_FromString(ret->services[0]->host));
  PyDict_SetItem(retdict,PyString_FromString("schema_version"),
                 PyString_FromString(ret->services[0]->schema_version));
  PyDict_SetItem(retdict,PyString_FromString("tier"),
                 PyString_FromString(ret->services[0]->tier));

  return retdict;

}

/* method table */

static PyMethodDef comm_lib_methods[] = {
  {"SDRegister",comm_lib_sdregister,METH_VARARGS},
  {"SDRemove",comm_lib_sdremove,METH_VARARGS},
  {"ServerInit",comm_lib_serverinit,METH_VARARGS},
  {"ClientInit",comm_lib_clientinit,METH_VARARGS},
  {"ServerClose",comm_lib_serverclose,METH_VARARGS},
  {"ClientClose",comm_lib_clientclose,METH_VARARGS},
  {"SendMessage",comm_lib_sendmessage,METH_VARARGS},
  {"RecvMessage",comm_lib_recvmessage,METH_VARARGS},
  {"SendRecv",comm_lib_sendrecv,METH_VARARGS},
  {"GetConfig",get_config_dict,METH_VARARGS},
  {"SetServerProtocol",set_servprot,METH_VARARGS},
  {"StaticLookup",static_lookup,METH_VARARGS},
  {NULL,NULL}
};

/* method getter */
static PyObject *comm_lib_getattr(PyObject *self,char *attr){
  return Py_FindMethod(comm_lib_methods,self,attr);
}

/* destructor */
static PyObject *comm_lib_close(PyObject *self,PyObject *args){
  SSS sss;
  /*pthread_mutex_t *mutex;*/

  sss = ((comm_lib *)self)->sss;
  /*mutex = &((comm_lib *)self)->mutex;*/

  /*pthread_mutex_lock(mutex);*/

  SSS_Close(sss);

  /*pthread_mutex_unlock(mutex);*/

  /*pthread_mutex_destroy(mutex);*/

  return Py_None;
}

/* type definition */

static PyTypeObject SSS_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "comm_lib",
  sizeof(comm_lib),
  0,
  (destructor)comm_lib_close,
  0,
  (getattrfunc)comm_lib_getattr,
};

/* python constructor */
static PyObject *comm_lib_new(PyObject *self,PyObject *args,PyObject *keywords){
  char *conffile;
  int debug;
  PyObject *result = NULL;

  static char *kwlist[] = {"conffile","debug",NULL};

  conffile = NULL;
  debug = 0;
  
  if (!PyArg_ParseTupleAndKeywords(args,keywords,"|si",kwlist,&conffile,&debug))
    return NULL;

  result = comm_lib_NEW(conffile,debug);

  return result;
}

/* module methods */

static PyMethodDef methods[] = {
  {"comm_lib",(PyCFunction)comm_lib_new,METH_VARARGS|METH_KEYWORDS},
  {NULL,NULL}
};

/* module init function */

void initssslib(void){
  PyObject *module,*dict;
  
  module = Py_InitModule("ssslib",methods);
  dict = PyModule_GetDict(module);

  SSS_ParseError = PyErr_NewException("ssslib.ParseError", NULL, NULL);
  PyDict_SetItemString(dict, "ParseError", SSS_ParseError);
  SSS_ProtocolError = PyErr_NewException("ssslib.ProtocolError", NULL, NULL);
  PyDict_SetItemString(dict, "ProtocolError", SSS_ProtocolError);
  SSS_SocketError = PyErr_NewException("ssslib.SocketError", NULL, NULL);
  PyDict_SetItemString(dict, "SocketError", SSS_SocketError);
  SSS_ConnectError = PyErr_NewException("ssslib.ConnectError", NULL, NULL);
  PyDict_SetItemString(dict, "ConnectError", SSS_ConnectError);
  SSS_CommunicationError = PyErr_NewException("ssslib.CommuncationError", NULL, NULL);
  PyDict_SetItemString(dict, "CommunicationError", SSS_CommunicationError);
  SSS_InternalError = PyErr_NewException("ssslib.InternalError", NULL, NULL);
  PyDict_SetItemString(dict, "InternalError", SSS_InternalError);
  SSS_InitError = PyErr_NewException("ssslib.InitError",NULL,NULL);
  PyDict_SetItemString(dict, "InitError", SSS_InitError);
  SSS_LookupError = PyErr_NewException("ssslib.LookupError",NULL,NULL);
  PyDict_SetItemString(dict, "LookupError", SSS_LookupError);
  SSS_UnhandledError = PyErr_NewException("ssslib.UnhandledError",NULL,NULL);
  PyDict_SetItemString(dict, "UnhandledError", SSS_UnhandledError);
  
}

